# -*- coding: utf-8 -*-
#
# Example of a very simple constraint extension module.
#
# Extension modules are used to extend TDDA Discovery and Verification
# to other data stores/types etc. --- so there are 'extensions' for
# Pandas DataFrames and various databases (PostgreSQL, MySQL, SQLite and
# MongoDB) included with the TDDA library.
#
# This file illustrates how to build a simple extension, treating the
# file system as a data source and metadata about the files (name, size)
# as the 'fields' to be checked in a directory (which takes the place
# of a table or dataset).
#
# USAGE:
#
# To use this, you need to set the environment variable TDDA_EXTENSIONS
# to include the fully-qualified class defined in here with a command.
#
# If this file is stored as /home/ann/constraints_examples/files_extension.py,
# use
#
#  export TDDA_EXTENSIONS=constraints_examples.files_extension.TDDAFilesExtension
#
# and make your PYTHONPATH is such that this is importable (i.e. that the
# PYTHONPATH includes /home/ann in this case). You can check this by trying
#
#     from constraints_examples.files_extension import TDDAFilesExtension
#
# in a Python REPL.
#
# With TDDA_EXTENSIONS and PYTHONPATH set appropriately, the "tdda" command
# should now include this extension for using constraints on filesystem
# directories.
#
# To generate a set of constraints in the current directory, and put the
# generated constraints into a file in /tmp, use:
#
#   tdda discover . /tmp/directoryconstraints.tdda
#
# To verify some other directory against those constraints, use:
#
#   tdda verify some_other_directory /tmp/directoryconstraints.tdda
#
# To detect files that do not conform to the constraints, as a CSV file, use:
#
#   tdda detect some_other_directory /tmp/directoryconstraints.tdda bad.csv
#

from __future__ import print_function

import csv
import os
import re
import sys
from collections import OrderedDict

from tdda.constraints.flags import discover_parser, discover_flags
from tdda.constraints.flags import verify_parser, verify_flags
from tdda.constraints.flags import detect_parser, detect_flags
from tdda.constraints.extension import ExtensionBase

from tdda.constraints.base import (
    DatasetConstraints,
    Detection,
)
from tdda.constraints.baseconstraints import (
    BaseConstraintCalculator,
    BaseConstraintVerifier,
    BaseConstraintDetector,
    BaseConstraintDiscoverer,
)
from tdda.rexpy import rexpy


class TDDAFilesExtension(ExtensionBase):
    def __init__(self, argv, verbose=False):
        ExtensionBase.__init__(self, argv, verbose=verbose)

    def applicable(self):
        return any(os.path.isdir(a) for a in self.argv)

    def help(self, stream=sys.stdout):
        print('  - Filesystem directories', file=stream)

    def spec(self):
        return 'a directory or folder'

    def discover(self):
        parser = discover_parser()
        parser.add_argument('directory', nargs=1, help='directory path')
        parser.add_argument('constraints', nargs='?',
                            help='name of constraints file to create')
        params = {}
        flags = discover_flags(parser, self.argv[1:], params)
        params['path'] = flags.directory[0] if flags.directory else None
        params['constraints_path'] = (flags.constraints if flags.constraints
                                      else None)
        constraints = discover_directory(**params)
        results = constraints.to_json()
        if params['constraints_path']:
            with open(params['constraints_path'], 'w') as f:
                f.write(results)
        else:
            print(results)
        return results

    def verify(self):
        parser = verify_parser()
        parser.add_argument('directory', nargs=1, help='directory path')
        parser.add_argument('constraints', nargs=1,
                            help='constraints file to verify against')
        return self.verify_or_detect(parser, detect=False)

    def detect(self):
        parser = detect_parser()
        parser.add_argument('directory', nargs=1, help='directory path')
        parser.add_argument('constraints', nargs=1,
                            help='constraints file to verify against')
        parser.add_argument('outpath', nargs=1,
                            help='file to write detected records to')
        return self.verify_or_detect(parser, detect=True)

    def verify_or_detect(self, parser, detect=False):
        params = {}
        if detect:
            flags = detect_flags(parser, self.argv[1:], params)
        else:
            flags = verify_flags(parser, self.argv[1:], params)
        params['path'] = flags.directory[0] if flags.directory else None
        params['constraints_path'] = (flags.constraints[0] if flags.constraints
                                      else None)
        if detect:
            params['outpath'] = flags.outpath[0] if flags.outpath else None
            results = detect_directory_from_file(**params)
        else:
            results = verify_directory_from_file(**params)
        if self.verbose:
            print(results)
        return results


def discover_directory(path, constraints_path=None, **kwargs):
    disco = FilesConstraintDiscoverer(path, **kwargs)
    return disco.discover()


def verify_directory_from_file(path, constraints_path, **kwargs):
    fv = FilesConstraintVerifier(path, **kwargs)
    constraints = DatasetConstraints(loadpath=constraints_path)
    return fv.verify(constraints, **kwargs)


def detect_directory_from_file(path, constraints_path, **kwargs):
    fv = FilesConstraintVerifier(path, **kwargs)
    constraints = DatasetConstraints(loadpath=constraints_path)
    return fv.detect(constraints, **kwargs)


class FilesConstraintCalculator(BaseConstraintCalculator):
    def __init__(self, path):
        self.path = path

    def get_column_names(self):
        return ['name', 'size']

    def get_nrecords(self):
        return len(os.listdir(self.path))

    def types_compatible(self, x, y, colname):
        return type(x) == type(y)

    def calc_min(self, colname):
        names = os.listdir(self.path)
        vals = (names if colname == 'name'
                else [os.path.getsize(self.fullpath(f)) for f in names])
        return min(vals)

    def calc_max(self, colname):
        names = os.listdir(self.path)
        vals = (names if colname == 'name'
                else [os.path.getsize(self.fullpath(f)) for f in names])
        return max(vals)

    def calc_min_length(self, colname):
        assert colname == 'name'
        names = os.listdir(self.path)
        return min([len(f) for f in names])

    def calc_max_length(self, colname):
        assert colname == 'name'
        names = os.listdir(self.path)
        return max([len(f) for f in names])

    def calc_tdda_type(self, colname):
        return 'string' if colname == 'name' else 'int'

    def calc_null_count(self, colname):
        return 0

    def calc_non_null_count(self, colname):
        return len(os.listdir(self.path))

    def calc_nunique(self, colname):
        names = os.listdir(self.path)
        vals = (names if colname == 'name'
                else [os.path.getsize(self.fullpath(f)) for f in names])
        uniques = set(vals)
        return len(uniques)

    def calc_unique_values(self, colname, include_nulls=True):
        names = os.listdir(self.path)
        vals = (names if colname == 'name'
                else [os.path.getsize(self.fullpath(f)) for f in names])
        uniques = set(vals)
        return list(uniques)

    def calc_non_integer_values_count(self, colname):
        raise Exception('files should not require non_integer_values_count')

    def calc_all_non_nulls_boolean(self, colname):
        raise Exception('files should not require all_non_nulls_boolean')

    def find_rexes(self, colname, values=None):
        assert colname == 'name'
        if not values:
            values = calc_unique_values(colname)
        return rexpy.extract(values)

    def calc_rex_constraint(self, colname, constraint, detect=False):
        # note that this should return violations, so None means success
        assert colname == 'name'
        names = os.listdir(self.path)
        for f in names:
            for r in constraint.value:
                if re.match(r, f) is not None:
                    break
            else:
                return True
        return None   # should really return a Detection object

    def fullpath(self, f):
        return os.path.join(self.path, f)


class FilesConstraintDetector(BaseConstraintDetector):
    def __init__(self, path):
        self.path = path
        self.names = os.listdir(self.path)
        self.sizes = [os.path.getsize(self.fullpath(f)) for f in self.names]
        self.min_ok = {}
        self.max_ok = {}
        self.min_length_ok = {}
        self.max_length_ok = {}
        self.no_dups_ok = {}
        self.values_ok = {}
        self.rex_ok = {}

    def detect_min_constraint(self, colname, value, precision, epsilon):
        vals = self.names if colname == 'name' else self.sizes
        self.min_ok[colname] = [v >= value for v in vals]

    def detect_max_constraint(self, colname, value, precision, epsilon):
        vals = self.names if colname == 'name' else self.sizes
        self.max_ok[colname] = [v <= value for v in vals]

    def detect_min_length_constraint(self, colname, value):
        assert colname == 'name'
        self.min_length_ok[colname] = [len(v) >= value for v in self.names]

    def detect_max_length_constraint(self, colname, value):
        assert colname == 'name'
        self.max_length_ok[colname] = [len(v) <= value for v in self.names]

    def detect_no_duplicates_constraint(self, colname, value):
        vals = self.names if colname == 'name' else self.sizes
        multiplicity = [len([c for c in vals if c == v]) for v in vals]
        self.no_dups_ok[colname] = [m <= value
                                    for m in multiplicity]

    def detect_allowed_values_constraint(self, colname, allowed_values,
                                         violations):
        vals = self.names if colname == 'name' else self.sizes
        self.values_ok[colname] = [v not in violations for v in vals]

    def detect_rex_constraint(self, colname, violations):
        assert colname == 'name'
        self.rex_ok[colname] = [v not in violations for v in vals]

    def write_detected_records(self,
                               detect_outpath=None,
                               detect_write_all=False,
                               detect_per_constraint=False,
                               detect_output_fields=None,
                               detect_index=False,
                               detect_in_place=False,
                               **kwargs):
        input_fields = ['name', 'size']
        if detect_outpath:
            if detect_output_fields is None:
                detect_output_fields = input_fields
            else:
                for k in detect_output_fields:
                    if k not in ('name', 'size'):
                        raise Exception('Unknown column %s' % k)
            cnames = ('min', 'max', 'min_length', 'max_length',
                      'no_dups', 'values', 'rex')
            cvalues = (self.min_ok, self.max_ok, self.min_length_ok,
                       self.max_length_ok, self.no_dups_ok,
                       self.values_ok, self.rex_ok)
            ok_output_names = ['%s_%s_ok' % (fname, cname)
                               for fname in input_fields
                               for cname, v in zip(cnames, cvalues)
                                   if fname in v]
            bad_output_names = []
            if len(detect_output_fields) == 0:
                bad_output_names.append('RowNumber')
            for k in detect_output_fields:
                bad_output_names.append(k)
            for k in ok_output_names:
                bad_output_names.append(k)

            with open(detect_outpath, 'w') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=bad_output_names)
                writer.writeheader()
                for i, (name, size) in enumerate(zip(self.names, self.sizes)):
                    record = OrderedDict()
                    if len(detect_output_fields) == 0:
                        record['RowNumber'] = i + 1
                    for k in detect_output_fields:
                        record[k] = name if k == 'name' else size
                    bad = False
                    for field in input_fields:
                        for constraint, ok in (zip(cnames, cvalues)):
                            if field in ok:
                                ok_field_name = '%s_%s_ok' % (field, constraint)
                                record[ok_field_name] = ok[field][i]
                                if ok[field][i] is False:
                                    bad = True
                    if bad or detect_write_all:
                        writer.writerow(record)
        return None

    def fullpath(self, f):
        return os.path.join(self.path, f)


class FilesConstraintDiscoverer(FilesConstraintCalculator,
                                BaseConstraintDiscoverer):
    def __init__(self, path, **kwargs):
        FilesConstraintCalculator.__init__(self, path)
        BaseConstraintDiscoverer.__init__(self, **kwargs)


class FilesConstraintVerifier(FilesConstraintCalculator,
                              FilesConstraintDetector,
                              BaseConstraintVerifier):
    def __init__(self, path, **kwargs):
        FilesConstraintCalculator.__init__(self, path)
        FilesConstraintDetector.__init__(self, path)
        BaseConstraintVerifier.__init__(self, **kwargs)





# -*- coding: utf-8 -*-

"""
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import sys
import tempfile

from tdda.referencetest.checkpandas import PandasComparison
from tdda.referencetest.checkfiles import FilesComparison


# DEFAULT_FAIL_DIR is the default location for writing failing output
# if assertStringCorrect or assertFileCorrect fail with 'preprocessing'
# in place. This can be overridden using the set_defaults() class method.
DEFAULT_FAIL_DIR = os.environ.get('TDDA_FAIL_DIR', tempfile.gettempdir())


def tag(test):
    """
    Decorator for tests, so that you can specify you only want to
    run a tagged subset of tests, with the -1 or --tagged option.
    """
    test._tagged = True
    return test


class ReferenceTest(object):
    """
    The :py:class:`~tdda.referencetest.referencetest.ReferenceTest` class
    provides support for comparing results against a set of reference
    "known to be correct" results.

    The functionality provided by this class can be used with:

        - the standard Python :py:mod:`unittest` framework, using the
          :py:class:`~tdda.referencetest.referencetestcase.ReferenceTestCase`
          class. This is a subclass of, and therefore a drop-in replacement
          for, :py:class:`unittest.TestCase`. It extends that class with all
          of the methods from the
          :py:class:`~tdda.referencetest.referencetest.ReferenceTest` class.

        - the :py:mod:`pytest` framework, using the
          :py:mod:`~tdda.referencetest.referencepytest` module.
          This module provides all of the methods from the
          :py:class:`~tdda.referencetest.referencetest.ReferenceTest` class,
          exposed as functions that can be called directly from tests
          in a :py:mod:`pytest` suite.

    In addition to the various test-assertion methods, the module also
    provides some useful instance variables. All of these can be set
    explicitly in test setup code, using the :py:meth:`set_defaults`
    class method:

        *tmp_dir*
            The location where temporary files can be written to. It defaults
            to a system-specific temporary area.

        *print_fn*
            The function to use to display information while running tests,
            which should have the same signature as Python3's
            standard print function (the ``__future__`` print function
            in Python2).

        *verbose*
            Boolean verbose flag, to control reporting of errors while running
            tests. Reference tests tend to take longer to run than traditional
            unit tests, so it is often useful to be able to see
            information from failing tests as they happen, rather
            than waiting for the full report at the end.
    """

    # Verbose flag
    verbose = True

    # Temporary directory
    tmp_dir = DEFAULT_FAIL_DIR

    # Dictionary describing which kinds of reference files should be
    # regenerated when the tests are run. This should be set using the
    # set_regeneration() class-method. Can be initialized via the -w option.
    regenerate = {}

    # Dictionary describing default location for reference data, for
    # each kind. Can be initialized by set_default_data_location().
    default_data_locations = {}

    @classmethod
    def set_defaults(cls, **kwargs):
        """Set default parameters, at the class level. These defaults will
        apply to all instances of the class.

        The following parameters can be set:

            *verbose*:
                Sets the boolean verbose flag globally, to control
                reporting of errors while running tests. Reference
                tests tend to take longer to run than traditional
                unit tests, so it is often useful to be able to see
                information from failing tests as they happen, rather
                than waiting for the full report at the end. Verbose
                is set to ``True`` by default.

            *print_fn*: Sets the print function globally, to specify
                the function to use to display information while
                running tests.  The function have the same signature
                as Python3's standard print function (the
                ``__future__`` print function in Python2), a default
                print function is used which writes unbuffered to
                ``sys.stdout``.

            *tmp_dir*:
                Sets the tmp_dir property globally, to specify the
                directory where temporary files are written.
                Temporary files are created whenever a text file
                check fails and a 'preprocess' function has been
                specified. It's useful to be able to see the contents
                of the files after preprocessing has taken place,
                so preprocessed versions of the files are written
                to this directory, and their pathnames are included
                in the failure messages. If not explicitly set by
                :py:meth:`set_defaults()`, the environment variable
                *TDDA_FAIL_DIR* is used, or, if that is not defined,
                it defaults to */tmp*, *c:\\temp* or whatever
                :py:func:`tempfile.gettempdir()` returns, as
                appropriate.

        """
        for k in kwargs:
            if k == 'verbose':
                cls.verbose = kwargs[k]
            elif k == 'print_fn':
                cls.print_fn = kwargs[k]
            elif k == 'tmp_dir':
                cls.tmp_dir = kwargs[k]
            else:
                raise Exception('set_defaults: Unrecogized option %s' % k)

    @classmethod
    def set_regeneration(cls, kind=None, regenerate=True):
        """
        Set the regeneration flag for a particular kind of reference file,
        globally, for all instances of the class.

        If the regenerate flag is set to ``True``, then the framework will
        regenerate reference data of that kind, rather than comparing.

        All of the regeneration flags are set to False by default.
        """
        cls.regenerate[kind] = regenerate

    @classmethod
    def set_default_data_location(cls, location, kind=None):
        """
        Declare the default filesystem location for reference files of a
        particular kind. This sets the location globally, and will apply
        to all instances of the class.

        The instance method :py:meth:`set_data_location()` can be used to set
        the per-kind data locations for an individual instance of the class.

        If calls to :py:meth:`assertFileCorrect()` (etc) are made for
        kinds of reference data that hasn't had their location defined
        explicitly, then the
        default location is used. This is the location declared for
        the ``None`` *kind* and this default **must** be specified.

        If you haven't even defined the ``None`` default, and you make calls
        to :py:meth:`assertFileCorrect()` (etc) using relative pathnames for
        the reference data files, then it can't check correctness, so it will
        raise an exception.

        """
        cls.default_data_locations[kind] = os.path.normpath(location)

    def __init__(self, assert_fn):
        """
        Initializer for a ReferenceTest instance.

            *assert_fn*:
                Function to be used to make assertions for
                unit-tests. It should take two parameters:

                    - a value (which should evaluate as ``True`` for
                      the test to pass)
                    - a string (to report details of how a test
                      failed, if the value does not evaluate as ``True``).
        """
        self.assert_fn = assert_fn
        self.reference_data_locations = dict(self.default_data_locations)
        self.pandas = PandasComparison(print_fn=self.print_fn,
                                       verbose=self.verbose)
        self.files = FilesComparison(print_fn=self.print_fn,
                                     verbose=self.verbose,
                                     tmp_dir=self.tmp_dir)

    def all_fields_except(self, exclusions):
        """
        Helper function, for using with *check_data*, *check_types* and
        *check_order* parameters to assertion functions for Pandas DataFrames.
        It returns the names of all of the fields in the DataFrame being
        checked, apart from the ones given.

        *exclusions* is a list of field names.
        """
        return self.pandas.all_fields_except(exclusions)

    def set_data_location(self, location, kind=None):
        """
        Declare the filesystem location for reference files of a
        particular kind. Typically you would subclass
        `ReferenceTestCase` and pass in these locations though its
        __init__ method when constructing an instance of
        ReferenceTestCase as a superclass.

        If calls to :py:meth:`assertFileCorrect()` (etc) are made for
        kinds of reference data that hasn't had their location defined
        explicitly, then the
        default location is used. This is the location declared for
        the ``None`` *kind* and this default **must** be specified.

        This method overrides any global defaults set from calls to the
        set_default_data_location class-method.

        If you haven't even defined the ``None`` default, and you make calls
        to :py:meth:`assertFileCorrect()` (etc) using relative pathnames for
        the reference data files, then it can't check correctness, so it will
        raise an exception.

        """
        self.reference_data_locations[kind] = os.path.normpath(location)

    def assertDataFramesEqual(self, df, ref_df,
                              actual_path=None, expected_path=None,
                              check_data=None, check_types=None,
                              check_order=None, condition=None, sortby=None,
                              precision=None):
        """Check that an in-memory Pandas `DataFrame` matches an in-memory
        reference one.

            *df*:
                Actual `DataFrame`.

            *ref_df*:
                Expected `DataFrame`.

            *actual_path*:
                (Optional) path for file where
                actual DataFrame originated, used for error messages.

            *expected_path*:
                (Optional) path for file where
                expected DataFrame originated, used for error messages.

            *check_data*:
                (Optional) restriction of fields whose values should
                be compared.
                Possible values are:

                    - ``None`` or ``True`` (to apply the comparison to
                      *all* fields; this is the default).
                    - ``False`` (to skip the comparison completely)
                    - a list of field names (to check only these fields)
                    - a function taking a ``DataFrame`` as its single
                      parameter,
                      and returning a list of field names to check.

            *check_types*:
                (Optional) restriction of fields whose types should be
                compared.
                See *check_data* (above) for possible values.

            *check_order*:
                (Optional) restriction of fields whose (relative)
                order should be compared.
                See *check_data* (above) for possible values.

            *check_extra_cols*:
                (Optional) restriction of extra fields in the actual dataset
                which, if found, will cause the check to fail.
                See *check_data* (above) for possible values.

            *sortby*:
                (Optional) specification of fields to sort by before comparing.

                    - ``None`` or ``False`` (do not sort; this is the default)
                    - ``True`` (to sort on all fields based on their
                      order in the reference datasets; you probably
                      don't want to use this option)
                    - a list of field names (to sort on these fields, in order)
                    - a function taking a ``DataFrame`` (which will be
                      the reference data frame) as its single
                      parameter,
                      and returning a list of field names to sort on.

            *condition*:
                (Optional) filter to be applied to datasets before comparing.
                It can be ``None``, or can be a function that
                takes a `DataFrame` as its single parameter and
                returns a vector of booleans (to specify which rows
                should be compared).

            *precision*:
                (Optional) number of decimal places to use for
                floating-point comparisons.  Default is not to perform
                rounding.

        Raises :py:class:`NotImplementedError` if Pandas is not available.

        """
        r = self.pandas.check_dataframe(df, ref_df,
                                        actual_path=actual_path,
                                        expected_path=expected_path,
                                        check_data=check_data,
                                        check_types=check_types,
                                        check_order=check_order,
                                        condition=condition,
                                        sortby=sortby,
                                        precision=precision)
        (failures, msgs) = r
        self._check_failures(failures, msgs)

    def assertDataFrameCorrect(self, df, ref_csv, actual_path=None,
                               kind='csv', csv_read_fn=None,
                               check_data=None, check_types=None,
                               check_order=None, condition=None, sortby=None,
                               precision=None, **kwargs):
        """Check that an in-memory Pandas DataFrame matches a reference one
        from a saved reference CSV file.

            *df*:
                Actual DataFrame.

            *ref_csv*:
                Name of reference CSV file. The location of the
                reference file is determined by the configuration
                via :py:meth:`set_data_location()`.

            *actual_path*:
                Optional parameter, giving path for file where
                actual DataFrame originated, used for error
                messages.

            *kind*:
                (Optional) reference kind (a string; see above), used to locate
                the reference CSV file.

            *csv_read_fn*:
                (Optional) function to read a CSV file to obtain
                a pandas DataFrame. If ``None``, then a default
                CSV loader is used.

                The default CSV loader function is a wrapper around Pandas
                :py:func:`pd.read_csv()`, with default options as follows:

                    - ``index_col`` is ``None``
                    - ``infer_datetime_format`` is ``True``
                    - ``quotechar`` is ``"``
                    - ``quoting`` is :py:const:`csv.QUOTE_MINIMAL`
                    - ``escapechar`` is ``\\`` (backslash)
                    - ``na_values`` are the empty string, ``"NaN"``, and ``"NULL"``
                    - ``keep_default_na`` is ``False``

            *check_data*:
                (Optional) restriction of fields whose values should
                be compared.
                Possible values are:

                    - ``None`` or ``True`` (to apply the comparison to
                      *all* fields; this is the default).
                    - ``False`` (to skip the comparison completely)
                    - a list of field names (to check only these fields)
                    - a function taking a ``DataFrame`` as its single
                      parameter,
                      and returning a list of field names to check.

            *check_types*:
                (Optional) restriction of fields whose types should be
                compared.
                See *check_data* (above) for possible values.

            *check_order*:
                (Optional) restriction of fields whose (relative)
                order should be compared.
                See *check_data* (above) for possible values.

            *check_extra_cols*:
                (Optional) restriction of extra fields in the actual dataset
                which, if found, will cause the check to fail.
                See *check_data* (above) for possible values.

            *sortby*:
                (Optional) specification of fields to sort by before comparing.

                    - ``None`` or ``False`` (do not sort; this is the default)
                    - ``True`` (to sort on all fields based on their
                      order in the reference datasets; you probably
                      don't want to use this option)
                    - a list of field names (to sort on these fields, in order)
                    - a function taking a ``DataFrame`` (which will be
                      the reference data frame) as its single
                      parameter,
                      and returning a list of field names to sort on.

            *condition*:
                (Optional) filter to be applied to datasets before comparing.
                It can be ``None``, or can be a function that
                takes a `DataFrame` as its single parameter and
                returns a vector of booleans (to specify which rows
                should be compared).

            *precision*:
                (Optional) number of decimal places to use for
                floating-point comparisons.  Default is not to perform
                rounding.

        Raises :py:class:`NotImplementedError` if Pandas is not available.

        """
        expected_path = self._resolve_reference_path(ref_csv, kind=kind)
        if self._should_regenerate(kind):
            self._write_reference_dataset(df, expected_path)
        else:
            ref_df = self.pandas.load_csv(expected_path, loader=csv_read_fn)
            self.assertDataFramesEqual(df, ref_df,
                                       actual_path=actual_path,
                                       expected_path=expected_path,
                                       check_data=check_data,
                                       check_types=check_types,
                                       check_order=check_order,
                                       condition=condition,
                                       sortby=sortby,
                                       precision=precision)

    def assertCSVFileCorrect(self, actual_path, ref_csv,
                             kind='csv', csv_read_fn=None,
                             check_data=None, check_types=None,
                             check_order=None, condition=None, sortby=None,
                             precision=None, **kwargs):
        """Check that a CSV file matches a reference one.

            *actual_path*:
                Actual CSV file.

            *ref_csv*:
                Name of reference CSV file. The location of the
                reference file is determined by the configuration
                via :py:meth:`set_data_location()`.

            *kind*:
                (Optional) reference kind (a string; see above), used to locate
                the reference CSV file.

            *csv_read_fn*:
                (Optional) function to read a CSV file to obtain
                a pandas DataFrame. If ``None``, then a default
                CSV loader is used.

                The default CSV loader function is a wrapper around Pandas
                :py:func:`pd.read_csv()`, with default options as follows:

                    - ``index_col`` is ``None``
                    - ``infer_datetime_format`` is ``True``
                    - ``quotechar`` is ``"``
                    - ``quoting`` is :py:const:`csv.QUOTE_MINIMAL`
                    - ``escapechar`` is ``\\`` (backslash)
                    - ``na_values`` are the empty string, ``"NaN"``, and ``"NULL"``
                    - ``keep_default_na`` is ``False``

            *check_data*:
                (Optional) restriction of fields whose values should
                be compared.
                Possible values are:

                    - ``None`` or ``True`` (to apply the comparison to
                      *all* fields; this is the default).
                    - ``False`` (to skip the comparison completely)
                    - a list of field names (to check only these fields)
                    - a function taking a ``DataFrame`` as its single
                      parameter,
                      and returning a list of field names to check.

            *check_types*:
                (Optional) restriction of fields whose types should be
                compared.
                See *check_data* (above) for possible values.

            *check_order*:
                (Optional) restriction of fields whose (relative)
                order should be compared.
                See *check_data* (above) for possible values.

            *check_extra_cols*:
                (Optional) restriction of extra fields in the actual dataset
                which, if found, will cause the check to fail.
                See *check_data* (above) for possible values.

            *sortby*:
                (Optional) specification of fields to sort by before comparing.

                    - ``None`` or ``False`` (do not sort; this is the default)
                    - ``True`` (to sort on all fields based on their
                      order in the reference datasets; you probably
                      don't want to use this option)
                    - a list of field names (to sort on these fields, in order)
                    - a function taking a ``DataFrame`` (which will be
                      the reference data frame) as its single
                      parameter,
                      and returning a list of field names to sort on.

            *condition*:
                (Optional) filter to be applied to datasets before comparing.
                It can be ``None``, or can be a function that
                takes a `DataFrame` as its single parameter and
                returns a vector of booleans (to specify which rows
                should be compared).

            *precision*:
                (Optional) number of decimal places to use
                for floating-point comparisons.  Default is not to
                perform rounding.

            *\*\*kwargs*:
                Any additional named parameters are passed
                straight through to the *csv_read_fn* function.

        Raises :py:class:`NotImplementedError` if Pandas is not available.

        """
        expected_path = self._resolve_reference_path(ref_csv, kind=kind)
        if self._should_regenerate(kind):
            self._write_reference_file(actual_path, expected_path)
        else:
            r = self.pandas.check_csv_file(actual_path, expected_path,
                                           check_data=check_data,
                                           check_types=check_types,
                                           check_order=check_order,
                                           condition=condition,
                                           sortby=sortby,
                                           precision=precision)
            (failures, msgs) = r
            self._check_failures(failures, msgs)

    def assertCSVFilesCorrect(self, actual_paths, ref_csvs,
                              kind='csv', csv_read_fn=None,
                              check_data=None, check_types=None,
                              check_order=None, condition=None, sortby=None,
                              precision=None, **kwargs):
        """Check that a set of CSV files match corresponding reference ones.

            *actual_paths*:
                List of actual CSV files.

            *ref_csvs*:
                List of names of matching reference CSV files. The
                location of the reference files is determined by
                the configuration via :py:meth:`set_data_location()`.

            *kind*:
                (Optional) reference kind (a string; see above), used to locate
                the reference CSV file.

            *csv_read_fn*:
                (Optional) function to read a CSV file to obtain
                a pandas DataFrame. If ``None``, then a default
                CSV loader is used.

                The default CSV loader function is a wrapper around Pandas
                :py:func:`pd.read_csv()`, with default options as follows:

                    - ``index_col`` is ``None``
                    - ``infer_datetime_format`` is ``True``
                    - ``quotechar`` is ``"``
                    - ``quoting`` is :py:const:`csv.QUOTE_MINIMAL`
                    - ``escapechar`` is ``\\`` (backslash)
                    - ``na_values`` are the empty string, ``"NaN"``, and ``"NULL"``
                    - ``keep_default_na`` is ``False``

            *check_data*:
                (Optional) restriction of fields whose values should
                be compared.
                Possible values are:

                    - ``None`` or ``True`` (to apply the comparison to
                      *all* fields; this is the default).
                    - ``False`` (to skip the comparison completely)
                    - a list of field names (to check only these fields)
                    - a function taking a ``DataFrame`` as its single
                      parameter,
                      and returning a list of field names to check.

            *check_types*:
                (Optional) restriction of fields whose types should be
                compared.
                See *check_data* (above) for possible values.

            *check_order*:
                (Optional) restriction of fields whose (relative)
                order should be compared.
                See *check_data* (above) for possible values.

            *check_extra_cols*:
                (Optional) restriction of extra fields in the actual dataset
                which, if found, will cause the check to fail.
                See *check_data* (above) for possible values.

            *sortby*:
                (Optional) specification of fields to sort by before comparing.

                    - ``None`` or ``False`` (do not sort; this is the default)
                    - ``True`` (to sort on all fields based on their
                      order in the reference datasets; you probably
                      don't want to use this option)
                    - a list of field names (to sort on these fields, in order)
                    - a function taking a ``DataFrame`` (which will be
                      the reference data frame) as its single
                      parameter,
                      and returning a list of field names to sort on.

            *condition*:
                (Optional) filter to be applied to datasets before comparing.
                It can be ``None``, or can be a function that
                takes a ``DataFrame`` as its single parameter and
                returns a vector of booleans (to specify which rows
                should be compared).

            *precision*:
                (Optional) number of decimal places to use for
                floating-point comparisons.  Default is not to perform
                rounding.

            *\*\*kwargs*:
                Any additional named parameters are passed straight
                through to the *csv_read_fn* function.

        Raises :py:class:`NotImplementedError` if Pandas is not available.

        """
        expected_paths = self._resolve_reference_paths(ref_csvs, kind=kind)
        if self._should_regenerate(kind):
            self._write_reference_files(actual_paths, expected_paths)
        else:
            r = self.pandas.check_csv_files(actual_paths, expected_paths,
                                            check_data=check_data,
                                            check_types=check_types,
                                            check_order=check_order,
                                            condition=condition,
                                            sortby=sortby,
                                            precision=precision)
            (failures, msgs) = r
            self._check_failures(failures, msgs)

    def assertStringCorrect(self, string, ref_path, kind=None,
                            lstrip=False, rstrip=False,
                            ignore_substrings=None,
                            ignore_patterns=None, ignore_lines=None,
                            preprocess=None, max_permutation_cases=0):
        """
        Check that an in-memory string matches the contents from a reference
        text file.

            *string*:
                The actual string.

            *ref_path*:
                The name of the reference file. The
                location of the reference file is
                determined by the configuration via
                :py:meth:`set_data_location()`.

            *kind*:
                The reference *kind*, used to locate the reference file.

            *lstrip*:
                If set to ``True``, both strings are
                left-stripped before the comparison is carried out.
                Note: the stripping is on a per-line basis.

            *rstrip*:
                If set to ``True``, both strings are
                right-stripped before the comparison is carried out.
                Note: the stripping is on a per-line basis.

            *ignore_substrings*:
                An optional list of substrings; lines
                containing any of these substrings will be
                ignored in the comparison.

            *ignore_patterns*:
                An optional list of regular expressions;
                lines will be considered to be the same if
                they only differ in substrings that match
                one of these regular expressions.
                The expressions must not contain parenthesised groups, and
                should only include explicit anchors if they
                need to refer to the whole line.

            *ignore_lines*
                An optional list of substrings; lines
                containing any of these substrings will be
                completely removed before carrying out the
                comparison. This is the means by which you
                would exclude 'optional' content.

            *preprocess*:
                An optional function that takes a list of
                strings and preprocesses it in some way; this
                function will be applied to both the actual and expected.

            *max_permutation_cases*:
                An optional number specifying the maximum
                number of permutations allowed; if the actual
                and expected lists differ only in that their
                lines are permutations of each other, and
                the number of such permutations does not
                exceed this limit, then the two are considered to be identical.

        """
        expected_path = self._resolve_reference_path(ref_path, kind=kind)
        if self._should_regenerate(kind):
            self._write_reference_result(string, expected_path)
        else:
            ilc = ignore_substrings
            ip = ignore_patterns
            mpc = max_permutation_cases
            r = self.files.check_string_against_file(string, expected_path,
                                                     actual_path=None,
                                                     lstrip=lstrip,
                                                     rstrip=rstrip,
                                                     ignore_substrings=ilc,
                                                     ignore_patterns=ip,
                                                     ignore_lines=ignore_lines,
                                                     preprocess=preprocess,
                                                     max_permutation_cases=mpc)
            (failures, msgs) = r
            self._check_failures(failures, msgs)

    def assertFileCorrect(self, actual_path, ref_path, kind=None,
                          lstrip=False, rstrip=False,
                          ignore_substrings=None,
                          ignore_patterns=None, ignore_lines=None,
                          preprocess=None, max_permutation_cases=0):
        """
        Check that a file matches the contents from a reference text file.

            *actual_path*:
                A path for a text file.

            *ref_path*:
                The name of the reference file. The
                location of the reference file is determined by
                the configuration via
                :py:meth:`set_data_location()`.

            *kind*:
                The reference *kind*, used to locate the reference file.

            *lstrip*:
                If set to ``True``, lines are left-stripped
                before the comparison is carried out.

            *rstrip*:
                If set to ``True``, lines are right-stripped before
                the comparison is carried out.

            *ignore_substrings*:
                An optional list of substrings; lines
                containing any of these substrings will be
                ignored in the comparison.

            *ignore_patterns*:
                An optional list of regular expressions;
                lines will be considered to be the same if
                they only differ in substrings that match one
                of these regular expressions. The expressions
                must not contain parenthesised groups, and
                should only include explicit anchors if they
                need to refer to the whole line.

            *ignore_lines*
                An optional list of substrings; lines
                containing any of these substrings will be
                completely removed before carrying out the
                comparison. This is the means by which you
                would exclude 'optional' content.

            *preprocess*:
                An optional function that takes a list of
                strings and preprocesses it in some way; this
                function will be applied to both the actual and expected.

            *max_permutation_cases*:
                An optional number specifying the maximum
                number of permutations allowed; if the actual
                and expected lists differ only in that their
                lines are permutations of each other, and
                the number of such permutations does not
                exceed this limit, then the two are considered to be identical.

        This should be used for unstructured data such as logfiles, etc.
        For CSV files, use :py:meth:`assertCSVFileCorrect` instead.

        """
        expected_path = self._resolve_reference_path(ref_path, kind=kind)
        if self._should_regenerate(kind):
            self._write_reference_file(actual_path, expected_path)
        else:
            mpc = max_permutation_cases
            r = self.files.check_file(actual_path, expected_path,
                                      lstrip=lstrip, rstrip=rstrip,
                                      ignore_substrings=ignore_substrings,
                                      ignore_patterns=ignore_patterns,
                                      ignore_lines=ignore_lines,
                                      preprocess=preprocess,
                                      max_permutation_cases=mpc)
            (failures, msgs) = r
            self._check_failures(failures, msgs)

    def assertFilesCorrect(self, actual_paths, ref_paths, kind=None,
                           lstrip=False, rstrip=False,
                           ignore_substrings=None,
                           ignore_patterns=None, ignore_lines=None,
                           preprocess=None, max_permutation_cases=0):
        """
        Check that a collection of files matche the contents from
        matching collection of reference text files.

            *actual_paths*:
                A list of paths for text files.

            *ref_paths*:
                A list of names of the matching reference
                files.  The location of the reference files
                is determined by the configuration via
                :py:meth:`set_data_location()`.

            *kind*:
                The reference *kind*, used to locate the reference files.
                All the files must be of the same kind.

            *lstrip*:
                If set to ``True``, lines are left-stripped before
                the comparison is carried out.

            *rstrip*:
                If set to ``True``, lines are right-stripped before the
                comparison is carried out.

            *ignore_substrings*:
                An optional list of substrings; lines
                containing any of these substrings will be
                ignored in the comparison.

            *ignore_patterns*:
                An optional list of regular expressions;
                lines will be considered to be the same if
                they only differ in substrings that match one
                of these regular expressions. The expressions
                must not contain parenthesised groups, and
                should only include explicit anchors if they
                need to refer to the whole line.

            *ignore_lines*
                An optional list of substrings; lines
                containing any of these substrings will be
                completely removed before carrying out the
                comparison. This is the means by which you
                would exclude 'optional' content.

            *preprocess*:
                An optional function that takes a list of
                strings and preprocesses it in some way; this
                function will be applied to both the actual
                and expected.

            *max_permutation_cases*:
                An optional number specifying the maximum
                number of permutations allowed; if the actual
                and expected lists differ only in that their
                lines are permutations of each other, and
                the number of such permutations does not
                exceed this limit, then the two are considered
                to be identical.

        This should be used for unstructured data such as logfiles, etc.
        For CSV files, use :py:meth:`assertCSVFileCorrect` instead.

        """
        expected_paths = self._resolve_reference_paths(ref_paths, kind=kind)
        if self._should_regenerate(kind):
            self._write_reference_files(actual_paths, expected_paths)
        else:
            mpc = max_permutation_cases
            r = self.files.check_files(actual_paths, expected_paths,
                                       lstrip=lstrip, rstrip=rstrip,
                                       ignore_substrings=ignore_substrings,
                                       ignore_patterns=ignore_patterns,
                                       ignore_lines=ignore_lines,
                                       preprocess=preprocess,
                                       max_permutation_cases=mpc)
            (failures, msgs) = r
            self._check_failures(failures, msgs)

    def _resolve_reference_path(self, path, kind=None):
        """
        Internal method for deciding where a reference data file should
        be looked for, if it has been specified using a relative path.
        """
        if self.reference_data_locations and not os.path.isabs(path):
            if kind not in self.reference_data_locations:
                kind = None
            if kind in self.reference_data_locations:
                path = os.path.join(self.reference_data_locations[kind], path)
            else:
                raise Exception('No reference data location for "%s"' % kind)
        return path

    def _resolve_reference_paths(self, paths, kind=None):
        """
        Internal method for resolving a list of reference data files,
        all of the same kind.
        """
        return [self._resolve_reference_path(p, kind=kind) for p in paths]

    def _should_regenerate(self, kind):
        """
        Internal method to determine if a particular kind of file
        should be regenerated.
        """
        if kind not in self.regenerate:
            kind = None
        return kind in self.regenerate and self.regenerate[kind]

    def _write_reference_file(self, actual_path, reference_path):
        """
        Internal method for regenerating reference data.
        """
        with open(actual_path) as fin:
            actual = fin.read()
        self._write_reference_result(actual, reference_path)

    def _write_reference_files(self, actual_paths, reference_paths):
        """
        Internal method for regenerating reference data for a list of
        files.
        """
        for (actual_path, expected_path) in zip(actual_paths, reference_paths):
            self._write_reference_file(actual_path, expected_path)

    def _write_reference_dataset(self, df, reference_path):
        """
        Internal method for regenerating reference data for a Pandas dataset
        """
        self.pandas.write_csv(df, reference_path)

    def _write_reference_result(self, result, reference_path):
        """
        Internal method for regenerating reference data from in-memory
        results.
        """
        with open(reference_path, 'w') as fout:
            fout.write(result)
        if self.verbose and self.print_fn:
            self.print_fn('Written %s' % reference_path)

    def _check_failures(self, failures, msgs):
        """
        Internal method for check for failures and reporting them.
        """
        self.assert_fn(failures == 0, '\n'.join(msgs))

    @staticmethod
    def _default_print_fn(*args, **kwargs):
        # Sometimes the framework needs to print messages. By default, it
        # will use this print function, but you can override it by passing
        # in a print_fn parameter to __init__.
        print(*args, **kwargs)
        outfile = kwargs.get('file', sys.stdout)
        outfile.flush()

    # Default print function
    print_fn = _default_print_fn


# Magic so that an instance of this class can masquerade as a module,
# so that all of its methods can be made available as top-level functions,
# to work will with frameworks like pytest.
ReferenceTest.__all__ = dir(ReferenceTest)




# -*- coding: utf-8 -*-

"""
TDDA constraint discovery and verification, common underlying functionality.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import re
import sys

from collections import OrderedDict

from tdda.constraints.base import (
    PRECISIONS,
    SIGNS,
    STANDARD_FIELD_CONSTRAINTS,
    verify, detect,
    native_definite,
    DatasetConstraints,
    FieldConstraints,
    Verification,
    TypeConstraint,
    MinConstraint, MaxConstraint, SignConstraint,
    MinLengthConstraint, MaxLengthConstraint,
    NoDuplicatesConstraint, MaxNullsConstraint,
    AllowedValuesConstraint, RexConstraint,
    EPSILON_DEFAULT,
    fuzzy_greater_than, fuzzy_less_than
)

from tdda.constraints.extension import (BaseConstraintCalculator,
                                        BaseConstraintDetector)

if sys.version_info[0] >= 3:
    unicode_string = str
    byte_string = bytes
else:
    unicode_string = unicode
    byte_string = str

DEBUG = False

TYPE_CHECKING_OPTIONS = ('strict', 'sloppy')
DEFAULT_TYPE_CHECKING = 'sloppy'

MAX_CATEGORIES = 20     # String fields with up to 20 categories will
                        # generate AllowedValues constraints


class BaseConstraintVerifier(BaseConstraintCalculator, BaseConstraintDetector):
    """
    The :py:mod:`BaseConstraintVerifier` class provides a generic
    framework for verifying constraints.

    A concrete implementation of this class is constructed by creating
    a mix-in subclass which inherits both from :py:mod:`BaseConstraintVerifier`
    and from specific implementations of :py:mod:`BaseConstraintCalculator`
    and :py:mod:`BaseConstraintDetector`.
    """
    def __init__(self, epsilon=None, type_checking=None, **kwargs):
        self.epsilon = EPSILON_DEFAULT if epsilon is None else epsilon
        self.type_checking = type_checking or DEFAULT_TYPE_CHECKING
        assert self.type_checking in TYPE_CHECKING_OPTIONS
        self.cache = {}

    def verifiers(self):
        """
        Returns a dictionary mapping constraint types to their callable
        (bound) verification methods.
        """
        return {
            'type': self.verify_tdda_type_constraint,
            'min': self.verify_min_constraint,
            'max': self.verify_max_constraint,
            'min_length': self.verify_min_length_constraint,
            'max_length': self.verify_max_length_constraint,
            'sign': self.verify_sign_constraint,
            'max_nulls': self.verify_max_nulls_constraint,
            'no_duplicates': self.verify_no_duplicates_constraint,
            'allowed_values': self.verify_allowed_values_constraint,
            'rex': self.verify_rex_constraint,
        }

    def verify(self, constraints, VerificationClass=Verification, **kwargs):
        """
        Apply verifiers to a set of constraints, for reporting
        """
        return verify(constraints, self.get_column_names(), self.verifiers(),
                      VerificationClass=VerificationClass,
                      detected_records_writer=self.write_detected_records,
                      **kwargs)

    def detect(self, constraints, VerificationClass=Verification,
               outpath=None, write_all=False, per_constraint=False,
               output_fields=None, index=False, in_place=False,
               rownumber_is_index=True, boolean_ints=False, **kwargs):
        """
        Apply verifiers to a set of constraints, for detection
        """
        return detect(constraints, self.get_column_names(), self.verifiers(),
                      VerificationClass=VerificationClass,
                      detect_outpath=outpath, detect_write_all=write_all,
                      detect_per_constraint=per_constraint,
                      detect_output_fields=output_fields,
                      detect_index=index,
                      detect_in_place=in_place,
                      detected_records_writer=self.write_detected_records,
                      rownumber_is_index=rownumber_is_index,
                      boolean_ints=boolean_ints,
                      **kwargs)

    def get_cached_value(self, value, colname, f):
        """
        Return cached value of colname, calculating it and caching it
        first, if it is not already there.
        """
        col_cache = self.cache_values(colname)
        if not value in col_cache:
            col_cache[value] = f(colname)
        return col_cache[value]

    def cache_values(self, colname):
        """
        Returns the dictionary for colname from the cache, first creating
        it if there isn't one on entry.
        """
        if not colname in self.cache:
            self.cache[colname] = {}
        return self.cache[colname]

    def verify_min_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the minimum value
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        precision = getattr(constraint, 'precision', 'fuzzy') or 'fuzzy'
        assert precision in PRECISIONS

        if self.is_null(value):   # a null minimum is not considered to be an
            return True           # active constraint, so is always satisfied

        m = self.get_min(colname)
        if self.is_null(m):       # If there are no values, no value can
            return True           # the minimum constraint

        if isinstance(value, datetime.datetime):
            m = self.to_datetime(m)

        if not self.types_compatible(m, value, colname):
            result = False
        elif precision == 'closed':
            result = m >= value
        elif precision == 'open':
            result = m > value
        else:
            result = fuzzy_greater_than(m, value, self.epsilon)

        if detect and result is False:
            self.detect_min_constraint(colname, value, precision, self.epsilon)
        return result

    def verify_max_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the maximum value
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        precision = getattr(constraint, 'precision', 'fuzzy') or 'fuzzy'
        assert precision in PRECISIONS

        if self.is_null(value):   # a null maximum is not considered to be an
            return True           # active constraint, so is always satisfied

        M = self.get_max(colname)
        if self.is_null(M):       # If there are no values, no value can
            return True           # the maximum constraint

        if isinstance(value, datetime.datetime):
            M = self.to_datetime(M)

        if not self.types_compatible(M, value, colname):
            result = False

        elif precision == 'closed':
            result = M <= value
        elif precision == 'open':
            result = M < value
        else:
            result = fuzzy_less_than(M, value, self.epsilon)

        if detect and result is False:
            self.detect_max_constraint(colname, value, precision, self.epsilon)
        return result

    def verify_min_length_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given (string) column satisfies the minimum length
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null minimum length is not considered
            return True           # to be an active constraint, so is always
                                  # satisfied

        m = self.get_min_length(colname)
        if self.is_null(m):       # If there are no values, no value can
            return True           # the minimum length constraint

        result = m >= value

        if detect and result is False:
            self.detect_min_length_constraint(colname, value)
        return result

    def verify_max_length_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given (string) column satisfies the maximum length
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null minimum length is not considered
            return True           # to be an active constraint, so is always
                                  # satisfied

        M = self.get_max_length(colname)
        if self.is_null(M):       # If there are no values, no value can
            return True           # the maximum length constraint

        result = M <= value

        if detect and result is False:
            self.detect_max_length_constraint(colname, value)
        return result

    def verify_tdda_type_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the supplied type constraint.
        """
        if not self.column_exists(colname):
            return False

        required_type = constraint.value
        allowed_types = (required_type if type(required_type) in (list, tuple)
                         else [required_type])
        if len(allowed_types) == 1 and self.is_null(allowed_types[0]):
            return True  # a null type is not considered to be an
                         # active constraint, so is always satisfied

        actual_type = self.get_tdda_type(colname)
        if self.type_checking == 'strict':
            result = actual_type in allowed_types
        else:
            if actual_type in allowed_types:
                result = True       # definitely OK of the types actually match
            elif 'int' in allowed_types and actual_type == 'real':
                result = self.get_non_integer_values_count(colname) == 0
            elif 'bool' in allowed_types and actual_type == 'real':
                result = self.get_non_integer_values_count(colname) == 0
            elif 'bool' in allowed_types and actual_type == 'string':
                # boolean columns with nulls get converted to dtype
                # object, which is usually used for strings
                result =  self.get_all_non_nulls_boolean(colname)
            else:
                result = False

        if detect and result is False:
            self.detect_tdda_type_constraint(colname, required_type)
        return result

    def verify_sign_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the supplied sign constraint.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null value (as opposed to the string
                                  # 'null') is not considered to be an
            return True           # active constraint, so is always satisfied

        m = self.get_min(colname)
        M = self.get_max(colname)
        if self.is_null(m):
            return True  # no values: cannot violate constraint

        if value == 'null':
             result = False
        elif value == 'positive':
            result = m > 0
        elif value == 'non-negative':
            result = m >= 0
        elif value == 'zero':
            result = m == M == 0
        elif value == 'non-positive':
            result = M <= 0
        elif value == 'negative':
            result = M < 0
        assert value in SIGNS

        if detect and result is False:
            self.detect_sign_constraint(colname, value)
        return result

    def verify_max_nulls_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the supplied constraint
        that it should contain no nulls.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null value is not considered to be an
            return True           # active constraint, so is always satisfied
        result = self.get_null_count(colname) <= value

        if result is True or not detect:
            return result
        self.detect_max_nulls_constraint(colname, value)
        return False

    def verify_no_duplicates_constraint(self, colname, constraint,
                                        detect=False):
        """
        Verify whether a given column satisfies the constraint supplied,
        that it should contain no duplicate (non-null) values.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null value is not considered to be an
            return True           # active constraint, so is always satisfied

        assert value == True      # value not really used; but should be True

        non_nulls = self.get_non_null_count(colname)
        result = self.get_nunique(colname) == non_nulls

        if detect and result is False:
            self.detect_no_duplicates_constraint(colname, value)
        return result

    def verify_allowed_values_constraint(self, colname, constraint,
                                         detect=False):
        """
        Verify whether a given column satisfies the constraint on allowed
        (string) values provided.
        """
        if not self.column_exists(colname):
            return False

        exclusions = self.allowed_values_exclusions()
        allowed_values = constraint.value
        if allowed_values is None:      # a null value is not considered
            return True                 # to be an active constraint,
                                        # so is always satisfied

        n_allowed_values = len(allowed_values)
        n_actual_values = self.get_nunique(colname)

        if not detect and n_actual_values > n_allowed_values:
            # can know the result without actually identifying values
            result = False
        else:
            actual_values = self.get_unique_values(colname)
            exclusions = exclusions or []

            violations = (set(actual_values) - set(allowed_values)
                                             - set(exclusions))
            result = len(violations) == 0

        if detect and result is False:
            self.detect_allowed_values_constraint(colname, allowed_values,
                                                  violations)
        return result

    def verify_rex_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies a given regular
        expression constraint (by matching at least one of the regular
        expressions given).
        """
        violations = self.calc_rex_constraint(colname, constraint,
                                              detect=detect)
        if bool(violations):
            # a truthy result means some values failed the constraint
            if detect:
                self.detect_rex_constraint(colname, violations)
            return False
        else:
            return True

    def get_min(self, colname):
        """Looks up cached minimum of column, or calculates and caches it"""
        return self.get_cached_value('min', colname, self.calc_min)

    def get_max(self, colname):
        """Looks up cached maximum of column, or calculates and caches it"""
        return self.get_cached_value('max', colname, self.calc_max)

    def get_min_length(self, colname):
        """
        Looks up cached minimum string length in column,
        or calculates and caches it
        """
        return self.get_cached_value('min_length', colname,
                                     self.calc_min_length)

    def get_max_length(self, colname):
        """
        Looks up cached maximum string length in column,
        or calculates and caches it
        """
        return self.get_cached_value('max_length', colname,
                                     self.calc_max_length)

    def get_tdda_type(self, colname):
        """
        Looks up cached tdda type of a column,
        or calculates and caches it
        """
        return self.get_cached_value('tdda_type', colname, self.calc_tdda_type)

    def get_null_count(self, colname):
        """
        Looks up or caches the number of nulls in a column,
        or calculates and caches it
        """
        return self.get_cached_value('null_count', colname,
                                     self.calc_null_count)

    def get_non_null_count(self, colname):
        """
        Looks up or caches the number of non-null values in a column,
        or calculates and caches it
        """
        return self.get_cached_value('non_null_count', colname,
                                     self.calc_non_null_count)

    def get_nunique(self, colname):
        """
        Looks up or caches the number of unique (distinct) values in a column,
        or calculates and caches it.
        """
        return self.get_cached_value('nunique', colname, self.calc_nunique)

    def get_unique_values(self, colname):
        """
        Looks up or caches the list of unique (distinct) values in a column,
        or calculates and caches it.
        """
        return self.get_cached_value('uniques', colname,
                                     self.calc_unique_values)

    def get_non_integer_values_count(self, colname):
        """
        Looks up or caches the number of non-integer values in a real column,
        or calculates and caches it.
        """
        return self.get_cached_value('non_integer_values_count', colname,
                                     self.calc_non_integer_values_count)

    def get_all_non_nulls_boolean(self, colname):
        """
        Looks up or caches the number of non-integer values in a real column,
        or calculates and caches it.
        """
        return self.get_cached_value('all_non_nulls_boolean', colname,
                                     self.calc_all_non_nulls_boolean)


class BaseConstraintDiscoverer(BaseConstraintCalculator):
    """
    The :py:mod:`BaseConstraintDiscoverer` class provides a generic
    framework for discovering constraints.

    A concrete implementation of this class is constructed by creating
    a mix-in subclass which inherits both from :py:mod:`BaseConstraintDiscover`
    and from a specific implementation of :py:mod:`BaseConstraintCalculator`.
    """
    def __init__(self, inc_rex=False, **kwargs):
        self.inc_rex = inc_rex

    def discover(self):
        field_constraints = []
        for col in self.get_column_names():
            constraints = self.discover_field_constraints(col)
            if constraints:
                field_constraints.append(constraints)
        if field_constraints:
            return DatasetConstraints(field_constraints)
        else:
            return None

    def discover_field_constraints(self, fieldname):
        min_constraint = max_constraint = None
        min_length_constraint = max_length_constraint = None
        sign_constraint = no_duplicates_constraint = None
        max_nulls_constraint = allowed_values_constraint = None
        rex_constraint = None

        type_ = self.calc_tdda_type(fieldname)
        if type_ == 'other':
            return None         # Unrecognized or complex
        else:
            type_constraint = TypeConstraint(type_)
        length = self.get_nrecords()

        if length > 0:  # Things are not very interesting when there is no data
            nNull = self.calc_null_count(fieldname)
            nNonNull = self.calc_non_null_count(fieldname)
            assert nNull + nNonNull == length
            if nNull < 2:
                max_nulls_constraint = MaxNullsConstraint(nNull)

            # Useful info:
            uniqs = None
            n_unique = -1   # won't equal number of non-nulls later on
            if type_ in ('string', 'int'):
                n_unique = self.calc_nunique(fieldname)
                if type_ == 'string':
                    if n_unique <= MAX_CATEGORIES:
                        uniqs = self.calc_unique_values(fieldname,
                                                        include_nulls=False)
                    if uniqs:
                        avc = AllowedValuesConstraint(uniqs)
                        allowed_values_constraint = avc

            if nNonNull > 0:
                if type_ == 'string':
                    # We don't generate a min, max or sign constraints for
                    # strings. But we do generate min and max length
                    # constraints
                    if (uniqs is None and n_unique > 0):
                        # There were too many for us to have bothered getting
                        # them all before, but we need them now.
                        uniqs = self.calc_unique_values(fieldname,
                                                        include_nulls=False)
                    if uniqs:
                        if type(uniqs[0]) is unicode_string:
                            L = [len(v) for v in uniqs]
                        else:
                            L = [len(v.decode('UTF-8')) for v in uniqs]
                        m = min(L)
                        M = max(L)
                        min_length_constraint = MinLengthConstraint(m)
                        max_length_constraint = MaxLengthConstraint(M)
                else:
                    # Non-string fields all potentially get min and max values
                    m = self.calc_min(fieldname)
                    M = self.calc_max(fieldname)
                    if not self.is_null(m):
                        min_constraint = MinConstraint(m)
                    if not self.is_null(M):
                        max_constraint = MaxConstraint(M)

                    # Non-date fields potentially get a sign constraint too.
                    if min_constraint and max_constraint and type_ != 'date':
                        if m == M == 0:
                            sign_constraint = SignConstraint('zero')
                        elif m >= 0:
                            sign = 'positive' if m > 0 else 'non-negative'
                            sign_constraint = SignConstraint(sign)
                        elif M <= 0:
                            sign = 'negative' if M < 0 else 'non-positive'
                            sign_constraint = SignConstraint(sign)
                        # else:
                            # mixed
                    elif self.is_null(m) and type_ != 'date':
                        sign_constraint = SignConstraint('null')

            if n_unique == nNonNull and n_unique > 1 and type_ != 'real':
                no_duplicates_constraint = NoDuplicatesConstraint()

        if type_ == 'string' and self.inc_rex:
            rex_constraint = RexConstraint(self.find_rexes(fieldname,
                                                           values=uniqs))

        constraints = [c for c in [type_constraint,
                                   min_constraint, max_constraint,
                                   min_length_constraint, max_length_constraint,
                                   sign_constraint, max_nulls_constraint,
                                   no_duplicates_constraint,
                                   allowed_values_constraint,
                                   rex_constraint]
                         if c is not None]
        return FieldConstraints(fieldname, constraints)



from tdda import rexpy

corpus = ['123-AA-971', '12-DQ-802', '198-AA-045', '1-BA-834']
results = rexpy.extract(corpus)
print('Number of regular expressions found: %d' % len(results))
for rex in results:
        print('   ' + rex)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys
import unittest

from tdda.constraints.testbase import *
from tdda.referencetest import ReferenceTestCase

try:
    from tdda.constraints.pd.testpdconstraints import *
except ImportError:
    print('Skipping Pandas tests', file=sys.stderr)

try:
    from tdda.constraints.db.testdbconstraints import *
except ImportError:
    print('Skipping Database tests', file=sys.stderr)


if __name__ == '__main__':
    ReferenceTestCase.main()



"""
Run all TDDA tests
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import unittest

from tdda.constraints.testconstraints import *
from tdda.rexpy.testrexpy import *
from tdda.referencetest.tests.alltests import *


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-

"""
The ``tdda`` command-line utility provides built-in support for constraint
discovery and verification for tabular data stored in CSV files, Pandas
DataFrames saved in ``.feather`` files, and for a tables in a variety of
different databases.

The utility can be extended to provide support for constraint discovery
and verification for other kinds of data, via its Python extension framework.

The framework will automatically use any extension implementations that
have been declared using the ``TDDA_EXTENSIONS`` environment variable. This
should be set to a list of class names, for Python classes that extend the
:py:class:`ExtensionBase` base class.

The class names in the ``TDDA_EXTENSIONS`` environment variable should be
colon-separated for Unix systems, or semicolon-separated for Microsoft
Windows. To be usable, the classes must be accessible by Python (either
by being installed in Pythons standard module directory, or by being
included in the ``PYTHONPATH`` environment variable.

For example::

    export TDDA_EXTENSIONS="mytdda.MySpecialExtension"
    export PYTHONPATH="/my/python/sources:$PYTHONPATH"

With these in place, the ``tdda`` command will include constraint discovery
and verification using the ``MySpecialExtension`` implementation class
provided in the Python file ``/my/python/sources/mytdda.py``.

An  example of a simple extension is included with the set of standard
examples. See :ref:`examples`.

Extension Overview
------------------

An extension should provide:

 - an implementation (subclass) of :py:class:`ExtensionBase`, to
   provide a command-line interface, extending the ``tdda`` command
   to support a particular type of input data.

 - an implementation (subclass) of :py:class:`BaseConstraintCalculator`,
   to provide methods for computing individual constraint results.

 - an implementation (subclass) of :py:class:`BaseConstraintDetector`,
   to provide methods for generating detection results.


A typical implementation looks like::

    from tdda.constraints.flags import discover_parser, discover_flags
    from tdda.constraints.flags import verify_parser, verify_flags
    from tdda.constraints.flags import detect_parser, detect_flags
    from tdda.constraints.extension import ExtensionBase
    from tdda.constraints.base import DatasetConstraints, Detection
    from tdda.constraints.baseconstraints import (BaseConstraintCalculator,
                                                  BaseConstraintVerifier,
                                                  BaseConstraintDetector,
                                                  BaseConstraintDiscoverer)
    from tdda.rexpy import rexpy

    class MyExtension(ExtensionBase):
        def applicable(self):
            ...

        def help(self, stream=sys.stdout):
            print('...', file=stream)

        def spec(self):
            return '...'

        def discover(self):
            parser = discover_parser()
            parser.add_argument(...)
            params = {}
            flags = discover_flags(parser, self.argv[1:], params)
            data = ... get data source from flags ...
            discoverer = MyConstraintDiscoverer(data, **params)
            constraints = discoverer.discover()
            results = constraints.to_json()
            ... write constraints JSON to output file
            return results

        def verify(self):
            parser = verify_parser()
            parser.add_argument(...)
            params = {}
            flags = verify_flags(parser, self.argv[1:], params)
            data = ... get data source from flags ...
            verifier = MyConstraintVerifier(data, **params)
            constraints = DatasetConstraints(loadpath=...)
            results = verifier.verify(constraints)
            return results

        def detect(self):
            parser = detect_parser()
            parser.add_argument(...)
            params = {}
            flags = detect_flags(parser, self.argv[1:], params)
            data = ... get data source from flags ...
            detector = MyConstraintDetector(data, **params)
            constraints = DatasetConstraints(loadpath=...)
            results = detector.detect(constraints)
            return results

Extension API
-------------
"""

import sys


class ExtensionBase:
    """
    An extension must provide a class that is based on the
    :py:class:`ExtensionBase` class, providing implementations for its
    :py:meth:`applicable`, :py:meth:`help`, :py:meth:`discover` and
    :py:meth:`verify` methods.
    """
    def __init__(self, argv, verbose=False):
        """
        A subclass of :py:class:`ExtensionBase` should call its superclass
        :py:meth:`__init__` initialisation method with a list of argument
        strings (such as ``sys.path``).
        """
        self.argv = argv
        self.verbose = verbose

    def applicable(self):
        """
        The :py:meth:`applicable` method should return ``True`` if the
        :py:attr:`argv` property contains command-line parameters that
        can be used by this implementation.

        For example, if the extension can handle data stored in Excel
        ``.xlsx`` files, then its :py:meth:`applicable` method should
        return ``True`` if any of its parameters are filenames that have
        a ``.xlsx`` suffix.
        """
        return False

    def help(self, stream=sys.stdout):
        """
        help(self, stream=sys.stdout)
        The :py:meth:`help` method should document itself by writing
        lines to the given output stream.

        This is used by the ``tdda`` command's ``help`` option.
        """
        pass

    def spec(self):
        """
        The :py:meth:`spec` method should return a short one-line string
        describing, briefly, how to specify the input source.
        """
        return ''

    def discover(self):
        """
        The :py:meth:`discover` method should implement constraint
        discovery.

        It should use the ``self.argv`` variable to get whatever other
        optional or mandatory flags or parameters are required to specify
        the data from which constraints are to be discovered, and the name
        of the file to which the constraints are to be written.
        """
        pass

    def verify(self):
        """
        The :py:meth:`verify` method should implement constraint
        verification.

        It should read constraints from a ``.tdda`` file specified on
        the command line, and verify these constraints on the data
        specified.

        It should use the ``self.argv`` variable to get whatever other
        optional or mandatory flags or parameters are required to specify
        the data on which the constraints are to be verified.
        """
        pass

    def detect(self):
        """
        The :py:meth:`detect` method should implement constraint
        detection.

        It should read constraints from a ``.tdda`` file specified on
        the command line, and verify these constraints on the data
        specified, and produce detection output.

        It should use the ``self.argv`` variable to get whatever other
        optional or mandatory flags or parameters are required to specify
        the data on which the constraints are to be verified, where the
        output detection data should be written, and detection-specific
        flags.
        """
        pass


class BaseConstraintCalculator:
    """
    The :py:mod:`BaseConstraintCalculator` class defines a default or dummy
    implementation of all of the methods that are required in order
    to implement a constraint discoverer or verifier via subclasses of the
    base :py:mod:`BaseConstraintDiscoverer` and :py:mod:`BaseConstraintVerifier`
    classes.
    """
    def is_null(self, value):
        """
        Determine whether a value is null
        """
        return value is None

    def to_datetime(self, value):
        """
        Convert a value to a datetime
        """
        return value

    def column_exists(self, colname):
        """
        Returns whether this column exists in the dataset
        """
        return colname in self.get_column_names()

    def get_column_names(self):
        """
        Returns a list containing the names of all the columns
        """
        raise NotImplementedError('column_names')

    def get_nrecords(self):
        """
        Return total number of records
        """
        raise NotImplementedError('nrecords')

    def types_compatible(self, x, y, colname):
        """
        Determine whether the types of two values are compatible
        """
        raise NotImplementedError('types_compatible')

    def allowed_values_exclusions(self):
        """
        Get list of values to ignore when computing allowed values
        """
        return [None]

    def calc_tdda_type(self, colname):
        """
        Calculates the TDDA type of a column
        """
        raise NotImplementedError('type')

    def calc_min(self, colname):
        """
        Calculates the minimum (non-null) value in the named column.
        """
        raise NotImplementedError('min')

    def calc_max(self, colname):
        """
        Calculates the maximum (non-null) value in the named column.
        """
        raise NotImplementedError('max')

    def calc_min_length(self, colname):
        """
        Calculates the length of the shortest string(s) in the named column.
        """
        raise NotImplementedError('min_length')

    def calc_max_length(self, colname):
        """
        Calculates the length of the longest string(s) in the named column.
        """
        raise NotImplementedError('max_length')

    def calc_null_count(self, colname):
        """
        Calculates the number of nulls in a column
        """
        raise NotImplementedError('null_count')

    def calc_non_null_count(self, colname):
        """
        Calculates the number of nulls in a column
        """
        raise NotImplementedError('non_null_count')

    def calc_nunique(self, colname):
        """
        Calculates the number of unique non-null values in a column
        """
        raise NotImplementedError('nunique')

    def calc_unique_values(self, colname, include_nulls=True):
        """
        Calculates the set of unique values (including or excluding nulls)
        in a column
        """
        raise NotImplementedError('unique_values')

    def calc_non_integer_values_count(self, colname):
        """
        Calculates the number of unique non-integer values in a column

        This is only required for implementations where a dataset column
        may contain values of mixed type.
        """
        raise NotImplementedError('non_integer_values_count')

    def calc_all_non_nulls_boolean(self, colname):
        """
        Checks whether all the non-null values in a column are boolean.
        Returns True of they are, and False otherwise.

        This is only required for implementations where a dataset column
        may contain values of mixed type.
        """
        raise NotImplementedError('all_non_nulls_boolean')

    def find_rexes(self, colname, values=None):
        """
        Generate a list of regular expressions that cover all of
        the patterns found in the (string) column.
        """
        raise NotImplementedError('find_rexes')

    def calc_rex_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies a given regular
        expression constraint (by matching at least one of the regular
        expressions given).

        Returns a 'truthy' value (typically the set of the strings that do
        not match any of the regular expressions) on failure, and a 'falsy'
        value (typically False or None or an empty set) if there are no
        failures. Any contents of the returned value are used in the case
        where detect is set, by the corresponding extension method for
        recording detection results.
        """
        raise NotImplementedError('verify_rex')


class BaseConstraintDetector:
    """
    The :py:mod:`BaseConstraintDetector` class defines a default or dummy
    implementation of all of the methods that are required in order
    to implement constraint detection via the a subclass of the base
    :py:mod:`BaseConstraintVerifier` class.
    """
    def detect_min_constraint(self, colname, value, precision, epsilon):
        """
        Detect failures for a min constraint.
        """
        pass

    def detect_max_constraint(self, colname, value, precision, epsilon):
        """
        Detect failures for a max constraint.
        """
        pass

    def detect_min_length_constraint(self, colname, value):
        """
        Detect failures for a min_length constraint.
        """
        pass

    def detect_max_length_constraint(self, colname, value):
        """
        Detect failures for a max_length constraint.
        """
        pass

    def detect_tdda_type_constraint(self, colname, value):
        """
        Detect failures for a type constraint.
        """
        pass

    def detect_sign_constraint(self, colname, value):
        """
        Detect failures for a sign constraint.
        """
        pass

    def detect_max_nulls_constraint(self, colname, value):
        """
        Detect failures for a max_nulls constraint.
        """
        pass

    def detect_no_duplicates_constraint(self, colname, value):
        """
        Detect failures for a no_duplicates constraint.
        """
        pass

    def detect_allowed_values_constraint(self, colname, value, violations):
        """
        Detect failures for an allowed_values constraint.
        """
        pass

    def detect_rex_constraint(self, colname, value, violations):
        """
        Detect failures for a rex constraint.
        """
        pass

    def write_detected_records(self,
                               detect_outpath=None,
                               detect_write_all=False,
                               detect_per_constraint=False,
                               detect_output_fields=None,
                               detect_index=False,
                               detect_in_place=False,
                               rownumber_is_index=True,
                               boolean_ints=False,
                               **kwargs):
        """
        Write out a detection dataset.

        Returns a :py:class:``~tdda.constraints.base.Detection`` object
        (or ``None``).
        """
        pass




# -*- coding: utf-8 -*-

"""
Extensions to the ``tdda`` command line tool, to support Pandas dataframes
and CSV files.
"""

from __future__ import print_function

import sys

from tdda.constraints.extension import ExtensionBase

from tdda.constraints.pd.discover import PandasDiscoverer
from tdda.constraints.pd.verify import PandasVerifier
from tdda.constraints.pd.detect import PandasDetector


class TDDAPandasExtension(ExtensionBase):
    def __init__(self, argv, verbose=False):
        ExtensionBase.__init__(self, argv, verbose=verbose)

    def applicable(self):
        for a in self.argv:
            if a.endswith('.csv') or a.endswith('.feather'):
                return True
        return False

    def help(self, stream=sys.stdout):
        print('  - Flat files (filename.csv)', file=stream)
        print('  - Pandas DataFrames (filename.feather)', file=stream)

    def spec(self):
        return 'a CSV file or a .feather file'

    def discover(self):
        return PandasDiscoverer(self.argv, verbose=self.verbose).discover()

    def verify(self):
        return PandasVerifier(self.argv, verbose=self.verbose).verify()

    def detect(self):
        return PandasDetector(self.argv, verbose=self.verbose).detect()




# -*- coding: utf-8 -*-
"""
The :py:mod:`tdda.referencetest` module includes a set of examples,
for both :py:mod:`unittest` and :py:mod:`pytest`.

To copy these examples to your own *referencetest-examples* subdirectory
(or to a location of your choice), run the command::

    tdda examples referencetest [mydirectory] 

Alternatively, you can copy all examples using the following command::
    
    tdda examples

which will create 3 separate sub-directories.
"""
from __future__ import absolute_import

from tdda import examples

if __name__ == '__main__':
    examples.copy_main('referencetest')




# -*- coding: utf-8 -*-

"""
Command line interface for constraint discovery and verification.

If pandas is available, constraints can be discovered and verified on
.csv files and saved .feather dataframe files.

If supported database drivers are available, constraints can be discovered
and verified on tables in databases.

Constraint discovery and verification may be available for other data
sources too, via any extensions specified in the `TDDA_EXTENSIONS`
environment variable, if these are loadable using the normal Python module
loading rules.
"""

from __future__ import print_function

import importlib
import os
import sys
import unittest

from tdda.examples import copy_examples
from tdda.constraints.pd.discover import pd_discover_parser
from tdda.constraints.pd.verify import pd_verify_parser
from tdda.constraints.pd.detect import pd_detect_parser

from tdda import __version__


HELP="""Use
    tdda discover      to perform constraint discovery
    tdda verify        to verify data against constraints
    tdda detect        to detect failed constraints on data
    tdda examples      to copy the example data and code
    tdda version       to print the TDDA version number
    tdda help          to print this help
    tdda help COMMAND  to print help on COMMAND (discover, verify or detect)
    tdda test          to run the tdda library's tests."""


STANDARD_EXTENSIONS = [
    'tdda.constraints.pd.extension.TDDAPandasExtension',
    'tdda.constraints.db.extension.TDDADatabaseExtension',
]


def help(extensions, cmd=None, stream=sys.stdout):
    if cmd:
        if cmd in ('discover', 'verify', 'detect'):
            # display detailed help for discover, verify or detect.
            # note that we use the Pandas variant to show the details, but
            # we also list the various input sources for all of the other
            # extensions (like databases), as providing the full detail
            # for everything would probably not be very helpful,
            print(file=stream)
            if cmd == 'discover':
                pd_discover_parser().print_help(stream)
            elif cmd == 'verify':
                pd_verify_parser().print_help(stream)
            elif cmd == 'detect':
                pd_detect_parser().print_help(stream)
            print('\n%s is available for the following:'
                  % cmd.title(), file=stream)
            for ext in extensions:
                ext.help(stream)
            print(file=stream)
        elif cmd == 'examples':
            print('\ntdda examples [module] [directory]\n\n'
                  'Write out example code and data for a particular module '
                  '(referencetest,\nconstraints or rexpy), to the specified '
                  'directory.\n'
                  '\nIf no module is specified, examples for all three are '
                  'written out.\n'
                  '\nIf no output directory is specified, the examples are '
                  'written to a subdirectory\nof the current directory.\n'
                  '\nTo write out all of the examples for all three modules to '
                  'subdirectories\nwithin the current directory, just use:\n'
                  '    tdda examples\n', file=stream)
        else:
            print('\nNo help available for %s. Try one of the following:\n'
                  '    tdda help discover\n'
                  '    tdda help verify\n'
                  '    tdda help detect\n'
                  '    tdda help examples\n')
    else:
        print(HELP, file=stream)
        print(file=stream)
        print('Constraint discovery and verification is available for:\n',
            file=stream)
        for ext in extensions:
            ext.help(stream=stream)
            print(file=stream)
        print('\nUse "tdda help COMMAND" to get more detailed help about'
              'a particular command.\nE.g. "tdda help verify"\n',
              file=stream)


def load_extension(ext):
    """
    Dynamically load an extension class, which must be available
    using the normal module loading rules. i.e., needs to be already
    availble via sys.path (through $PYTHONPATH, or otherwise).
    """
    components = ext.split('.')
    classname = components[-1]
    modulename = '.'.join(components[:-1])
    try:
        mod = importlib.import_module(modulename)
        return getattr(mod, classname, None)
    except ImportError as e:
        print('Warning: no tdda constraint module %s (%s)'
              % (modulename, str(e)), file=sys.stderr)
        return None


def load_all_extensions(argv, verbose=False):
    """
    Load all extensions specified via the TDDA_EXTENSIONS environment variable,
    and then load all the standard extensions.
    """
    extension_class_names = []
    if 'TDDA_EXTENSIONS' in os.environ:
        extension_class_names.extend(os.environ['TDDA_EXTENSIONS'].split(':'))
    extension_class_names.extend(STANDARD_EXTENSIONS)
    extension_classes = [load_extension(e) for e in extension_class_names]
    return [e(argv, verbose=verbose) for e in extension_classes if e]


def no_constraints(name, msg, argv, extensions):
    """
    When no constraint discovery or verification could be done, show
    some help about it.
    """
    inputs = [a for a in argv
                if not a.startswith('-') and not a.endswith('.tdda')]
    if inputs:
        print('%s for %s' % (msg, ' '.join(inputs)), file=sys.stderr)
    help(extensions, name, stream=sys.stderr)


def main_with_argv(argv, verbose=True):
    extensions = load_all_extensions(argv[1:], verbose=verbose)

    if len(argv) == 1:
        help(extensions, stream=sys.stderr)
        sys.exit(1)
    name = argv[1]

    if name in ('discover', 'disco'):
        for ext in extensions:
            if ext.applicable():
                return ext.discover()
        no_constraints(name, 'No discovery available', argv[2:], extensions)
    elif name == 'verify':
        for ext in extensions:
            if ext.applicable():
                return ext.verify()
        no_constraints(name, 'No verification available', argv[2:], extensions)
    elif name == 'detect':
        for ext in extensions:
            if ext.applicable():
                return ext.detect()
        no_constraints(name, 'No detection available', argv[2:], extensions)
    elif name == 'examples':
        item = argv[2] if len(argv) > 2 else '.'
        if item in ('referencetest', 'constraints', 'rexpy'):
            dest = argv[3] if len(argv) > 3 else '.'
            copy_examples(item, destination=dest, verbose=verbose)
        else:
            dest = argv[2] if len(argv) > 2 else '.'
            for item in ('referencetest', 'constraints', 'rexpy'):
                copy_examples(item, destination=dest, verbose=verbose)
    elif name in ('version', '-v', '--version'):
        print(__version__)
    elif name == 'test':
        sys.exit(os.system('%s -m tdda.testtdda' % sys.executable) != 0)
    elif name in ('help', '-h', '-?', '--help'):
        cmd = sys.argv[2] if len(sys.argv) > 2 else None
        help(extensions, cmd, stream=sys.stderr)
    else:
        help(extensions, stream=sys.stderr)
        sys.exit(1)


def main():
    main_with_argv(sys.argv)


if __name__ == '__main__':
    main()




import os
from setuptools import setup, find_packages

# version import is from tdda subdirectory here, not from some other install.
from tdda.version import version as __version__

def read(fname):
    # read contents of file
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def data(path, pathitems):
    # build list of additional files to package up from a subdirectory
    names = []
    for relpath in pathitems:
        subpath = path + [relpath]
        dirname = os.path.join(*subpath)
        for name in os.listdir(dirname):
            pathname = os.path.join(relpath, name)
            fullpathname = os.path.join(dirname, name)
            if os.path.isdir(fullpathname):
                names.extend(data(path, [pathname]))
            else:
                names.append(pathname)
    return names

setup(
    name='tdda',
    version=__version__,
    description='Test Driven Data Analysis',
    long_description=read('README.md'),
    author='Stochastic Solutions Limited',
    author_email='info@StochasticSolutions.com',
    license='MIT',
    url='http://www.stochasticsolutions.com',
    download_url='https://github.com/tdda/tdda',
    keywords='tdda constraint referencetest rexpy',
    packages=find_packages(),
    package_data={
        'tdda.referencetest': data(['tdda', 'referencetest'], ['examples']),
        'tdda.referencetest.tests': data(['tdda', 'referencetest', 'tests'],
                                         ['testdata']),
        'tdda.constraints': data(['tdda', 'constraints'],
                                 ['testdata', 'examples'])
                            + ['tdda_json_file_format.md'],
        'tdda.rexpy': data(['tdda', 'rexpy'], ['examples']),
        'tdda': ['README.md', 'LICENSE.txt'],
    },
    entry_points = {
        'console_scripts': [
            'tdda = tdda.constraints.console:main',
            'rexpy = tdda.rexpy.rexpy:main',
        ],
    },
    zip_safe=False,
)




# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import json
import math
import os
import time
import sys
import unittest

from collections import OrderedDict

from tdda.constraints.base import (
    DatasetConstraints,
    Fields,
    FieldConstraints,
    MinConstraint,
    MaxConstraint,
    SignConstraint,
    TypeConstraint,
    MaxNullsConstraint,
    NoDuplicatesConstraint,
    AllowedValuesConstraint,
    MinLengthConstraint,
    MaxLengthConstraint,
    constraint_class,
    strip_lines,
    sort_constraint_dict,
)

isPython2 = sys.version_info[0] < 3

TESTDATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                            'testdata')


class TestConstraints(unittest.TestCase):

    def test_constraint_repr(self):
        self.assertEqual(repr(MinConstraint(7)),
                         'MinConstraint(value=7, precision=None)')
        self.assertEqual(repr(MinConstraint('a')),
                         "MinConstraint(value='a', precision=None)")
        self.assertEqual(repr(MinConstraint('a', precision='closed')),
                         "MinConstraint(value='a', precision='closed')")
        self.assertEqual(repr(MinLengthConstraint(3)),
                         "MinLengthConstraint(value=3)")
        self.assertEqual(repr(MaxConstraint(-3)),
                         'MaxConstraint(value=-3, precision=None)')
        self.assertEqual(repr(MaxConstraint('KJ')),
                         "MaxConstraint(value='KJ', precision=None)")
        self.assertEqual(repr(MaxConstraint(4.2, precision='closed')),
                         "MaxConstraint(value=4.2, precision='closed')")
        self.assertEqual(repr(MaxLengthConstraint(0)),
                         "MaxLengthConstraint(value=0)")
        self.assertEqual(repr(SignConstraint('positive')),
                         "SignConstraint(value='positive')")
        self.assertEqual(repr(MaxNullsConstraint(0)),
                         "MaxNullsConstraint(value=0)")
        self.assertEqual(repr(NoDuplicatesConstraint()),
                         "NoDuplicatesConstraint(value=True)")
        self.assertEqual(repr(TypeConstraint('int')),
                         "TypeConstraint(value='int')")
        self.assertEqual(repr(TypeConstraint(['int', 'real'])),
                         "TypeConstraint(value=['int', 'real'])")
        self.assertEqual(repr(AllowedValuesConstraint(['a', 'b'])),
                         "AllowedValuesConstraint(value=['a', 'b'])")

    def test_constraint_class(self):
        goods = {
            'type': 'TypeConstraint',
            'min': 'MinConstraint',
            'min_length': 'MinLengthConstraint',
            'max': 'MaxConstraint',
            'max_length': 'MaxLengthConstraint',
            'sign': 'SignConstraint',
            'max_nulls': 'MaxNullsConstraint',
            'no_duplicates': 'NoDuplicatesConstraint',
            'allowed_values': 'AllowedValuesConstraint',
        }
        for k,v in goods.items():
            self.assertEqual(constraint_class(k), v)

    def testBadConstraints(self):
        if isPython2:
            self.assertRaisesRegexp(TypeError, 'unexpected keyword',
                                    SignConstraint, precision='closed')
        else:
            self.assertRaisesRegex(TypeError, 'unexpected keyword',
                                   SignConstraint, precision='closed')
        self.assertRaises(AssertionError,
                          MinConstraint, 3, precision='unknown')
        self.assertRaises(AssertionError,
                          SignConstraint, 'not too positive')
        self.assertRaises(AssertionError,
                          TypeConstraint, 'float')
        self.assertRaises(AssertionError,
                          TypeConstraint, ['int', 'float'])
        self.assertRaises(AssertionError,
                          TypeConstraint, ['int', None])

    def testFieldConstraintsDict(self):
        c = FieldConstraints('one', [TypeConstraint('int'),
                                           MinConstraint(3),
                                           MaxConstraint(7),
                                           SignConstraint('positive'),
                                           MaxNullsConstraint(0),
                                           NoDuplicatesConstraint()])
        dfc = Fields([c])
        self.assertEqual(strip_lines(json.dumps(dfc.to_dict_value(),
                                                indent=4)),
                         '''{
    "one": {
        "type": "int",
        "min": 3,
        "max": 7,
        "sign": "positive",
        "max_nulls": 0,
        "no_duplicates": true
    }
}''')

        c = FieldConstraints('one', [TypeConstraint('int'),
                                     MinConstraint(3, precision='closed'),
                                     MaxConstraint(7, precision='fuzzy'),
                                     SignConstraint('positive'),
                                     MaxNullsConstraint(0),
                                     NoDuplicatesConstraint()])
        dfc = Fields([c])
        self.assertEqual(strip_lines(json.dumps(dfc.to_dict_value(),
                                                indent=4)),
                         '''{
    "one": {
        "type": "int",
        "min": {
            "value": 3,
            "precision": "closed"
        },
        "max": {
            "value": 7,
            "precision": "fuzzy"
        },
        "sign": "positive",
        "max_nulls": 0,
        "no_duplicates": true
    }
}''')

    def testload(self):
        path = os.path.join(TESTDATA_DIR, 'ddd.tdda')
        constraints = DatasetConstraints(loadpath=path)
        constraints.sort_fields()
        actual = constraints.to_json()
        with open(path) as f:
            expected = json.dumps(sort_constraint_dict(json.loads(f.read())),
                                  indent=4) + '\n'
            self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
captureoutput.py: CaptureOutput
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys


class CaptureOutput(object):
    """
    Class for capturing a stream (typically) stdout.

    Typical Usage:

        c = CaptureOutput()  # for stdout; add stream='stderr' for stderr.
        <do stuff>
        c.restore()
        printed = str(c)
    """
    def __init__(self, echo=False, stream='stdout'):
        self.stream = stream
        if stream == 'stdout':
            self.saved = sys.stdout
            sys.stdout = self
        elif stream == 'stderr':
            self.saved = sys.stderr
            sys.stderr = self
        else:
            raise Exception('Unsupported capture stream %s' % stream)
        self.out = []
        self.echo = echo

    def write(self, s):
        self.out.append(s)
        if self.echo:
            self.saved.write(s)

    def flush(self):
        if self.echo:
            self.saved.flush()

    def restore(self):
        if self.stream == 'stdout':
            sys.stdout = self.saved
        else:
            sys.stderr = self.saved

    def __str__(self):
        return ''.join(self.out)





# -*- coding: utf-8 -*-

"""
Support for database constraint verification from the command-line tool
"""

from __future__ import division
from __future__ import print_function

USAGE = '''

Parameters:

  * table is one of:

     - a database table name
     - a schema-qualified table name of the form schema.table
     - a database table name qualified by database type, of the
       form dbtype:table or dbtype:schema.table

  * constraints.tdda is a JSON .tdda file constaining constraints.

'''

import argparse
import os
import sys

from tdda import __version__
from tdda.constraints.flags import verify_parser, verify_flags
from tdda.constraints.db.constraints import verify_db_table
from tdda.constraints.db.drivers import (database_connection, parse_table_name,
                                         database_arg_parser,
                                         database_arg_flags)


def verify_database_table_from_file(table, constraints_path,
                                    conn=None, dbtype=None, db=None,
                                    host=None, port=None, user=None,
                                    password=None, **kwargs):
    """
    Verify the given database table, against constraints in the .tdda
    file specified.

    Prints results to stdout.
    """
    (table, dbtype) = parse_table_name(table, dbtype)
    db = database_connection(table=table, conn=conn, dbtype=dbtype, db=db,
                             host=host, port=port,
                             user=user, password=password)
    print(verify_db_table(dbtype, db, table, constraints_path, **kwargs))


def get_verify_params(args):
    parser = database_arg_parser(verify_parser, USAGE)
    parser.add_argument('table', nargs=1, help='database table name')
    parser.add_argument('constraints', nargs=1,
                        help='constraints file to verify against')
    params = {}
    flags = database_arg_flags(verify_flags, parser, args, params)
    params['table'] = flags.table[0] if flags.table else None
    params['constraints_path'] = (flags.constraints[0] if flags.constraints
                                  else None)
    return params


class DatabaseVerifier:
    def __init__(self, argv, verbose=False):
        self.argv = argv
        self.verbose = verbose

    def verify(self):
        params = get_verify_params(self.argv[1:])
        verify_database_table_from_file(**params)


def main(argv):
    if len(argv) > 1 and argv[1] in ('-v', '--version'):
        print(__version__)
        sys.exit(0)
    v = DatabaseVerifier(argv)
    v.verify()


if __name__ == '__main__':
    main(sys.argv)

